from tensorflow import keras
from .generic_keras import _PlotLossesCallback


class PlotLossesCallback(_PlotLossesCallback, keras.callbacks.Callback):
    """Callback for tensorflow keras"""
    def __init__(self, **kwargs):
        keras.callbacks.Callback.__init__(self)
        _PlotLossesCallback.__init__(self, **kwargs)



from abc import ABC, abstractmethod

from livelossplot.main_logger import MainLogger


class BaseOutput(ABC):
    @abstractmethod
    def send(self, logger: MainLogger):
        """Abstract method - handle logs for a plugin"""
        ...

    def close(self):
        """Overwrite it with last steps"""
        ...

    def set_output_mode(self, mode: str):
        """Some of output plugins needs to know target format"""
        assert mode in ('notebook', 'script')
        self._set_output_mode(mode)

    def _set_output_mode(self, mode: str):
        """
        Args:
            mode: mode for callbacks - some of outputs need to change some behaviors,
                depending on the working environment (scripts and jupyter notebooks)
        """
        ...



# technical

from .base_output import BaseOutput

# default

from .matplotlib_plot import MatplotlibPlot
from .extrema_printer import ExtremaPrinter

# with external dependencies
# import are respective __init__ methods
# hack-ish, but works (and I am not aware of a more proper way to do so)

from .bokeh_plot import BokehPlot
from .neptune_logger import NeptuneLogger
from .tensorboard_logger import TensorboardLogger
from .tensorboard_tf_logger import TensorboardTFLogger

# with external dependencies

from . import matplotlib_subplots



from typing import List, Dict, Tuple

from livelossplot.main_logger import MainLogger, LogItem
from livelossplot.outputs.base_output import BaseOutput


class BokehPlot(BaseOutput):
    """Simple plugin for a bokeh framework"""
    def __init__(
        self,
        max_cols: int = 2,
        skip_first: int = 2,
        cell_size: Tuple[int, int] = (400, 300),
        output_file: str = './bokeh_output.html'
    ):
        """
        Args:
            max_cols: max number of charts in one row
            skip_first: flag, skip first log
            cell_size: size of one chart
            output_file: file to save the output
        """
        from bokeh import plotting, io, palettes
        self.plotting = plotting
        self.io = io
        self.plot_width, self.plot_height = cell_size
        self.max_cols = max_cols
        self.skip_first = skip_first  # think about it
        self.figures = {}
        self.is_notebook = False
        self.output_file = output_file
        self.colors = palettes.Category10[10]

    def send(self, logger: MainLogger) -> None:
        """Draw figures with metrics and show"""
        log_groups = logger.grouped_log_history()
        new_grid_plot = False
        for idx, (group_name, group_logs) in enumerate(log_groups.items(), start=1):
            fig = self.figures.get(group_name)
            if not fig:
                fig = self.plotting.figure(title=group_name)
                new_grid_plot = True
            self.figures[group_name] = self._draw_metric_subplot(fig, group_logs)
        if new_grid_plot:
            self._create_grid_plot()
        if self.is_notebook:
            self.io.push_notebook(handle=self.target)
        else:
            self.plotting.save(self.grid)

    def _draw_metric_subplot(self, fig, group_logs: Dict[str, List[LogItem]]):
        """
        Args:
            fig: bokeh Figure
            group_logs: groups with list of log items

        Notes:
            for now, with local imports, no output annotation  -> self.plotting.Figure
            there used to be skip first part, but I skip it first
        """
        from bokeh.models import ColumnDataSource, HoverTool
        for i, (name, logs) in enumerate(group_logs.items()):
            if len(logs) > 0:
                source = ColumnDataSource(
                    data={
                        'step': [log.step for log in logs],
                        'value': [log.value for log in logs],
                    }
                )
                fig.line(x='step', y='value', color=self.colors[i], legend_label=name, source=source)

        fig.add_tools(
            HoverTool(
                tooltips=[
                    ('step', '@step'),
                    ('value', '@value{0.3f}'),
                ],
                formatters={
                    'step': 'printf',
                    'value': 'printf',
                },
                mode='vline'
            )
        )
        return fig

    def _create_grid_plot(self):
        rows = []
        row = []
        for idx, fig in enumerate(self.figures.values(), start=1):
            row.append(fig)
            if idx % self.max_cols == 0:
                rows.append(row)
                row = []
        self.grid = self.plotting.gridplot(
            rows, sizing_mode='scale_width', plot_width=self.plot_width, plot_height=self.plot_height
        )
        self.target = self.plotting.show(self.grid, notebook_handle=self.is_notebook)

    def _set_output_mode(self, mode: str):
        """Set notebook or script mode"""
        self.is_notebook = mode == 'notebook'
        if self.is_notebook:
            self.io.output_notebook()
        else:
            self.io.output_file(self.output_file)



import warnings
from typing import Type, TypeVar, List, Union

import livelossplot
from livelossplot.main_logger import MainLogger
from livelossplot import outputs

BO = TypeVar('BO', bound=outputs.BaseOutput)


class PlotLosses:
    """
    Class collect metrics from the training engine and send it to plugins, when send is called
    """
    def __init__(
        self,
        outputs: List[Union[Type[BO], str]] = ['MatplotlibPlot', 'ExtremaPrinter'],
        mode: str = 'notebook',
        **kwargs
    ):
        """
        Args:
            outputs: list of output modules: objects inheriting from BaseOutput
                or strings for livelossplot built-in output methods with default parameters
            mode: Options: 'notebook' or 'script' - some of outputs need to change some behaviors,
                depending on the working environment
            **kwargs: key-arguments which are passed to MainLogger constructor
        """
        self.logger = MainLogger(**kwargs)
        self.outputs = [getattr(livelossplot.outputs, out)() if isinstance(out, str) else out for out in outputs]
        for out in self.outputs:
            out.set_output_mode(mode)

    def update(self, *args, **kwargs):
        """update logs with arguments that will be passed to main logger"""
        self.logger.update(*args, **kwargs)

    def send(self):
        """Method will send logs to every output class"""
        for output in self.outputs:
            output.send(self.logger)

    def draw(self):
        """Send method substitute from old livelossplot api"""
        warnings.warn('draw will be deprecated, please use send method', PendingDeprecationWarning)
        self.send()

    def reset_outputs(self) -> 'PlotLosses':
        """Resets all outputs.

        Returns:
            Plotlosses object (so it works for chaining)
        """
        self.outputs = []
        return self

    def to_matplotlib(self, **kwargs) -> 'PlotLosses':
        """Appends outputs.MatplotlibPlot output, with specified parameters.

        Args:
            **kwargs: keyword arguments for MatplotlibPlot

        Returns:
            Plotlosses object (so it works for chaining)
        """
        self.outputs.append(outputs.MatplotlibPlot(**kwargs))
        return self

    def to_extrema_printer(self, **kwargs) -> 'PlotLosses':
        """Appends outputs.ExtremaPrinter output, with specified parameters.

        Args:
            **kwargs: keyword arguments for ExtremaPrinter

        Returns:
            Plotlosses object (so it works for chaining)
        """
        self.outputs.append(outputs.ExtremaPrinter(**kwargs))
        return self

    def to_bokeh(self, **kwargs) -> 'PlotLosses':
        """Appends outputs.BokehPlot output, with specified parameters.

        Args:
            **kwargs: keyword arguments for BokehPlot

        Returns:
            Plotlosses object (so it works for chaining)
        """
        self.outputs.append(outputs.BokehPlot(**kwargs))
        return self

    def to_neptune(self, **kwargs) -> 'PlotLosses':
        """Appends outputs.NeptuneLogger output, with specified parameters.

        Args:
            **kwargs: keyword arguments for NeptuneLogger

        Returns:
            Plotlosses object (so it works for chaining)
        """
        self.outputs.append(outputs.NeptuneLogger(**kwargs))
        return self

    def to_tensorboard(self, **kwargs) -> 'PlotLosses':
        """Appends outputs.TensorboardLogger output, with specified parameters.

        Args:
            **kwargs: keyword arguments for TensorboardLogger

        Returns:
            Plotlosses object (so it works for chaining)
        """
        self.outputs.append(outputs.TensorboardLogger(**kwargs))
        return self

    def to_tensorboard_tf(self, **kwargs) -> 'PlotLosses':
        """Appends outputs.TensorboardTFLogger output, with specified parameters.

        Args:
            **kwargs: keyword arguments for TensorboardTFLogger

        Returns:
            Plotlosses object (so it works for chaining)
        """
        self.outputs.append(outputs.TensorboardTFLogger(**kwargs))
        return self



from typing import Dict, List

from livelossplot.main_logger import LogItem
from livelossplot.main_logger import MainLogger
from .base_output import BaseOutput


class ExtremaPrinter(BaseOutput):
    def __init__(
        self,
        massage_template: str = '\t{metric_name:16} \t (min: {min:8.3f},'
        ' max: {max:8.3f}, cur: {current:8.3f})'
    ):
        """
        Args:
            massage_template: you can specify massage which use all or a few values (min, max, current)
        """
        self.massage_template = massage_template
        self.extrema_cache = {}

    def send(self, logger: MainLogger):
        """Create massages with log_history and massage template"""
        log_groups = logger.grouped_log_history()
        massages = self._create_massages(log_groups)
        print('\n'.join(massages))

    def _create_massages(self, log_groups: Dict[str, Dict[str, List[LogItem]]]) -> List[str]:
        """Update cache and create massages"""
        massages = []
        for group_name, group_logs in log_groups.items():
            massages.append(group_name)
            for metric_name, metric_values in group_logs.items():
                self.update_extrema(metric_name, group_name, metric_values)
                msg = self.massage_template.format(
                    metric_name=metric_name, **self.extrema_cache[group_name][metric_name]
                )
                massages.append(msg)
        return massages

    def update_extrema(self, metric_name: str, group_name: str, metric_values: List[LogItem]) -> None:
        """Write highest, lower and current value to cache (or initialize if no exist)"""
        current_val = metric_values[-1].value
        if not self.extrema_cache.get(group_name):
            self.extrema_cache[group_name] = {}
        cache = self.extrema_cache[group_name].get(metric_name)
        if not cache:
            min_val = min(metric_values, key=lambda i: i.value).value
            max_val = max(metric_values, key=lambda i: i.value).value
            current_val = metric_values[-1].value
            self.extrema_cache[group_name][metric_name] = {
                'min': min_val,
                'max': max_val,
                'current': current_val,
            }
        else:
            min_val = min(cache['min'], current_val)
            max_val = max(cache['max'], current_val)
            cache['min'] = min_val
            cache['max'] = max_val
            cache['current'] = current_val

    @property
    def extrema_cache(self) -> Dict[str, Dict[str, Dict[str, float]]]:
        """Cache getter"""
        return self._extrema_cache

    @extrema_cache.setter
    def extrema_cache(self, value: Dict[str, Dict[str, Dict[str, float]]]) -> None:
        """Cache setter - can initialize cache only with empty dictionary"""
        if len(value) > 0:
            raise RuntimeError('Cannot overwrite cache with non empty dictionary')
        self._extrema_cache = value

    def close(self):
        """Clear cache"""
        self.extrema_cache = {}



from typing import Optional
from livelossplot.main_logger import MainLogger
from livelossplot.outputs.base_output import BaseOutput


class NeptuneLogger(BaseOutput):
    """See: https://github.com/neptune-ai/neptune-client
    YOUR_API_TOKEN and USERNAME/PROJECT_NAME
    """
    def __init__(self, api_token: Optional[str] = None, project_qualified_name: Optional[str] = None, **kwargs):
        """Set secrets and create experiment
        Args:
            api_token: your api token, you can create NEPTUNE_API_TOKEN environment variable instead
            project_qualified_name: <user>/<project>, you can create NEPTUNE_PROJECT environment variable instead
            **kwargs: keyword args, that will be passed to create_experiment function
        """
        import neptune
        self.neptune = neptune
        self.neptune.init(api_token=api_token, project_qualified_name=project_qualified_name)
        self.experiment = self.neptune.create_experiment(**kwargs)

    def close(self):
        """Close connection"""
        self.neptune.stop()

    def send(self, logger: MainLogger):
        """Send metrics collected in last step to neptune server"""
        for name, log_items in logger.log_history.items():
            last_log_item = log_items[-1]
            self.neptune.send_metric(name, x=last_log_item.step, y=last_log_item.value)



from typing import Dict

from poutyne.framework import Callback
from ..plot_losses import PlotLosses


class PlotLossesCallback(Callback):
    """Poutyne is a keras-like api framework for pytorch"""
    def __init__(self, **kwargs):
        """
        Args:
            **kwargs: keyword arguments that will be passed to PlotLosses constructor
        """
        super(PlotLossesCallback, self).__init__()
        self.liveplot = PlotLosses(**kwargs)
        self.metrics = None

    def on_train_begin(self, logs):
        """Create metric names"""
        metrics = ['loss'] + self.model.metrics_names
        self.metrics = list(metrics)
        self.metrics += ['val_' + metric for metric in metrics]

    def on_epoch_end(self, epoch: int, logs: Dict[str, float]):
        """Send metrics to livelossplot
        Args:
            epoch: epoch number
            logs: metrics with values
        """
        metric_logs = {metric: logs[metric] for metric in self.metrics if metric in logs}
        self.liveplot.update(metric_logs, epoch)
        self.liveplot.send()



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import ListedColormap


class BaseSubplot:
    def __init__(self):
        pass

    def draw(self, *args, **kwargs):
        raise Exception("Not implemented")

    def __call__(self, *args, **kwargs):
        self.draw(*args, **kwargs)


class LossSubplot(BaseSubplot):
    """To rewrire, this one now won't work"""
    def __init__(
        self, metric, title="", series_fmt={
            'training': '{}',
            'validation': 'val_{}'
        }, skip_first=2, max_epoch=None
    ):
        super().__init__(self)
        self.metric = metric
        self.title = title
        self.series_fmt = series_fmt
        self.skip_first = skip_first
        self.max_epoch = max_epoch
        raise NotImplementedError()

    def _how_many_to_skip(self, log_length, skip_first):
        if log_length < skip_first:
            return 0
        elif log_length < 2 * skip_first:
            return log_length - skip_first
        else:
            return skip_first

    def draw(self, logs):
        skip = self._how_many_to_skip(len(logs), self.skip_first)

        if self.max_epoch is not None:
            plt.xlim(1 + skip, self.max_epoch)

        for serie_label, serie_fmt in self.series_fmt.items():

            serie_metric_name = serie_fmt.format(self.metric)
            serie_metric_logs = [
                (log.get('_i', i + 1), log[serie_metric_name])
                for i, log in enumerate(logs[skip:]) if serie_metric_name in log
            ]

            if len(serie_metric_logs) > 0:
                xs, ys = zip(*serie_metric_logs)
                plt.plot(xs, ys, label=serie_label)

        plt.title(self.title)
        plt.xlabel('epoch')
        plt.legend(loc='center right')


class Plot1D(BaseSubplot):
    def __init__(self, model, X, Y):
        super().__init__(self)
        self.model = model
        self.X = X
        self.Y = Y

    def predict(self, model, X):
        # e.g. model(torch.fromnumpy(X)).detach().numpy()
        return model.predict(X)

    def draw(self, *args, **kwargs):
        plt.plot(self.X, self.Y, 'r.', label="Ground truth")
        plt.plot(self.X, self.predict(self.model, self.X), '-', label="Model")
        plt.title("Prediction")
        plt.legend(loc='lower right')


class Plot2d(BaseSubplot):
    def __init__(self, model, X, Y, valiation_data=(None, None), h=0.02, margin=0.25, device='cpu'):
        super().__init__()

        self.model = model
        self.X = X
        self.Y = Y
        self.X_test, self.Y_test = valiation_data

        # add size assertions

        self.cm_bg = plt.cm.RdBu
        self.cm_points = ListedColormap(['#FF0000', '#0000FF'])

        x_min = X[:, 0].min() - margin
        x_max = X[:, 0].max() + margin

        y_min = X[:, 1].min() - margin
        y_max = X[:, 1].max() + margin

        self.xx, self.yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))

        self.torch_device = device

    def _predict_pytorch(self, model, x_numpy):
        import torch
        x = torch.from_numpy(x_numpy).to(self.torch_device).float()
        return model(x).softmax(dim=1).detach().cpu().numpy()

    def predict(self, model, X):
        # e.g. model(torch.fromnumpy(X)).detach().numpy()
        return model.predict(X)

    def send(self, logger):
        Z = self._predict_pytorch(self.model, np.c_[self.xx.ravel(), self.yy.ravel()])[:, 1]
        Z = Z.reshape(self.xx.shape)
        plt.contourf(self.xx, self.yy, Z, cmap=self.cm_bg, alpha=.8)
        plt.scatter(self.X[:, 0], self.X[:, 1], c=self.Y, cmap=self.cm_points)
        if self.X_test is not None:
            plt.scatter(self.X_test[:, 0], self.X_test[:, 1], c=self.Y_test, cmap=self.cm_points, alpha=0.3)



from typing import Optional

import ignite.engine
from ignite.handlers import global_step_from_engine
from livelossplot.plot_losses import PlotLosses


class PlotLossesCallback:
    def __init__(self, train_engine: Optional[ignite.engine.Engine] = None, **kwargs):
        """
        Args:
            train_engine: engine with global step information, send metohod callback will be attached to it
                if None send method will be called on the end of each store call it may cause warnings and errors in
                the case of multiple engines attached

        Keyword Args:
            **kwargs: keyword args that will be passed to livelossplot PlotLosses class
        """
        self.liveplot = PlotLosses(**kwargs)
        self.train_engine = train_engine
        if self.train_engine:
            self.train_engine.add_event_handler(ignite.engine.Events.EPOCH_STARTED, self.send)
            self.train_engine.add_event_handler(ignite.engine.Events.COMPLETED, self.send)

    def attach(self, engine: ignite.engine.Engine):
        """Attach callback to ignite engine, attached method will be called on the end of each epoch
        Args:
            engine: engine that computes metrics on the end of each epoch and / or on the end of each iteration

        Notes:
            metrics computation plugins have to be attached before this one
        """
        engine.add_event_handler(ignite.engine.Events.EPOCH_COMPLETED, self.store)
        engine.add_event_handler(ignite.engine.Events.ITERATION_COMPLETED, self.store)

    def store(self, engine: ignite.engine.Engine):
        """Store computed metrics, that will be send to main logger
        Args:
            engine: engine with state.metrics
        """
        metrics = {}
        if not hasattr(engine.state, 'metrics') or len(engine.state.metrics) == 0:
            return
        kwargs = dict(
            current_step=global_step_from_engine(self.train_engine)
            (self.train_engine, self.train_engine.last_event_name)
        ) if self.train_engine else {}
        for key, val in engine.state.metrics.items():
            metrics[key] = val
        self.liveplot.update(metrics, **kwargs)
        if not self.train_engine:
            self.send()

    def send(self, _: Optional[ignite.engine.Engine] = None):
        self.liveplot.send()



from datetime import datetime
from os import path
from typing import Optional

from livelossplot.main_logger import MainLogger
from livelossplot.outputs.base_output import BaseOutput


class TensorboardLogger(BaseOutput):
    """
    Class write logs to TensorBoard (using pure TensorBoard, not one from TensorFlow).
    """
    def __init__(self, logdir: str = "./tensorboard_logs/", run_id: Optional[str] = None):
        """
        Args:
            logdir: dir where TensorBoard events will be written
            run_id: name for log id, otherwise it usses datetime
        """
        from tensorboard import summary
        self.summary = summary
        run_id = datetime.now().isoformat()[:-7].replace("T", " ").replace(":", "_") if run_id is None else run_id
        self._path = path.join(logdir, run_id)
        self.writer = summary.create_file_writer(self._path)

    def close(self):
        """Close tensorboard writer"""
        self.writer.close()

    def log_scalar(self, name: str, value: float, global_step: float):
        """
        Args:
            name: name of metric
            value: float value of metric
            global_step: current step of the training loop
        """
        with self.writer.as_default():
            self.summary.scalar(name, value, step=global_step)
        self.writer.flush()

    def send(self, logger: MainLogger):
        """Take log history from logger and store it in tensorboard event"""
        for name, log_items in logger.log_history.items():
            last_log_item = log_items[-1]
            self.log_scalar(name, last_log_item.value, last_log_item.step)



from setuptools import setup, find_packages
from os import path
import re


def readme():
    with open('README.md', encoding='utf-8') as f:
        return f.read()


setup(
    name='realtimeplt',
    version='0.0.1',
    python_requires=">3.5",
    install_requires=[
        'ipython', 'matplotlib;python_version>="3.6"', 'matplotlib<3.1;python_version<"3.6"',
        'numpy<1.18;python_version<"3.6"', 'bokeh;python_version>="3.6"', 'bokeh<=1.4.0;python_version<"3.6"'
    ],
    description='Realtime plot of loss plot in Jupyter Notebook for Keras, PyTorch and others.',
    long_description=readme(),
    long_description_content_type='text/markdown',
    
    author='Aryan Karn',
    author_email='aryankarnin@gmail.com',
    
    keywords=['keras', 'pytorch', 'plot', 'chart', 'deep-learning'],
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Jupyter',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Visualization',
        'License :: OSI Approved :: MIT License',
        
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=find_packages(),
    zip_safe=False
)



def PlotLossesKeras(**kwargs):
    """PlotLosses callback for Keras (as a standalone library, not a TensorFlow module).
    Args:
        **kwargs: key-arguments which are passed to PlotLosses

    Notes:
        Requires keras to be installed.
    """
    from .keras import PlotLossesCallback
    return PlotLossesCallback(**kwargs)


def PlotLossesKerasTF(**kwargs):
    """PlotLosses callback for Keras (as a module of TensorFlow).
    Args:
        **kwargs: key-arguments which are passed to PlotLosses

    Notes:
        Requires tensorflow to be installed.
    """
    from .tf_keras import PlotLossesCallback
    return PlotLossesCallback(**kwargs)


def PlotLossesPoutyne(**kwargs):
    """PlotLosses callback for Poutyne, a library for PyTorch.
    Args:
        **kwargs: key-arguments which are passed to PlotLosses

    Notes:
        Requires poutyne to be installed, <https://poutyne.org/>.
    """
    from .poutyne import PlotLossesCallback
    return PlotLossesCallback(**kwargs)


def PlotLossesIgnite(**kwargs):
    """PlotLosses callback for PyTorch-Ignite, a library for PyTorch.
    Args:
        **kwargs: key-arguments which are passed to PlotLosses

    Notes:
        Requires pytorch-ignite to be installed, <https://github.com/pytorch/ignite>.
    """
    from .pytorch_ignite import PlotLossesCallback
    return PlotLossesCallback(**kwargs)



"""
.. include:: ../README.md
"""

import sys
import warnings
from importlib.util import find_spec

from .main_logger import MainLogger
from .plot_losses import PlotLosses
from . import inputs
from .inputs import *
from . import outputs
from .version import __version__

_input_plugin_dict = {
    'keras': 'Keras',
    'tf_keras': 'KerasTF',
    'pytorch_ignite': 'Ignite',
    'poutyne': 'Poutyne',
}


class OldDependenciesFinder:
    """
    Data package module loader finder. This class sits on `sys.meta_path` and returns the
    loader it knows for a given path, if it knows a compatible loader.
    """
    @classmethod
    def find_spec(self, fullname: str, *_, **__):
        """This functions is what gets executed by the loader.
        Args:
            fullname: name of the called module
        """
        parts = fullname.split('.')
        if len(parts) == 2 and parts[0] == 'livelossplot' and parts[1] in _input_plugin_dict:
            name = parts[1]
            msg = 'livelossplot.{name} will be deprecated, please use livelossplot.inputs.{name}\n'
            msg += 'or use callback directly: from livelossplot import PlotLosses{new_name}'
            warnings.warn(msg.format(name=name, new_name=_input_plugin_dict[name]), DeprecationWarning)
            fullname = 'livelossplot.inputs.{name}'.format(name=name)
            return find_spec(fullname)
        return None


sys.meta_path.append(OldDependenciesFinder())

__all__ = [
    'MainLogger', 'inputs', 'outputs', 'PlotLosses', 'PlotLossesKeras', 'PlotLossesKerasTF', 'PlotLossesIgnite',
    'PlotLossesPoutyne'
]



import math
from typing import Tuple, List, Dict, Optional, Callable

import warnings

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from IPython.display import clear_output
from livelossplot.main_logger import MainLogger, LogItem
from livelossplot.outputs.base_output import BaseOutput


class MatplotlibPlot(BaseOutput):
    """NOTE: Removed figsize and dynamix_x_axis."""
    def __init__(
        self,
        cell_size: Tuple[int, int] = (6, 4),
        max_cols: int = 2,
        max_epoch: int = None,
        skip_first: int = 2,
        extra_plots: List[Callable[[MainLogger], None]] = [],
        figpath: Optional[str] = None,
        after_subplot: Optional[Callable[[plt.Axes, str, str], None]] = None,
        before_plots: Optional[Callable[[plt.Figure, np.ndarray, int], None]] = None,
        after_plots: Optional[Callable[[plt.Figure], None]] = None,
    ):
        """
        Args:
            cell_size: size of one chart
            max_cols: maximal number of charts in one row
            max_epoch: maximum epoch on x axis
            skip_first: number of first steps to skip
            extra_plots: extra charts functions
            figpath path: to save figure
            after_subplot: function which will be called after every subplot
            before_plots: function which will be called before all subplots
            after_plots: function which will be called after all subplots
        """
        self.cell_size = cell_size
        self.max_cols = max_cols
        self.skip_first = skip_first  # think about it
        self.extra_plots = extra_plots
        self.max_epoch = max_epoch
        self.figpath = figpath
        self.file_idx = 0  # now only for saving files
        self._after_subplot = after_subplot if after_subplot else self._default_after_subplot
        self._before_plots = before_plots if before_plots else self._default_before_plots
        self._after_plots = after_plots if after_plots else self._default_after_plots

    def send(self, logger: MainLogger):
        """Draw figures with metrics and show"""
        log_groups = logger.grouped_log_history()

        max_rows = math.ceil((len(log_groups) + len(self.extra_plots)) / self.max_cols)

        fig, axes = plt.subplots(max_rows, self.max_cols)
        axes = axes.reshape(-1, self.max_cols)
        self._before_plots(fig, axes, len(log_groups))

        for group_idx, (group_name, group_logs) in enumerate(log_groups.items()):
            ax = axes[group_idx // self.max_cols, group_idx % self.max_cols]
            self._draw_metric_subplot(ax, group_logs, group_name=group_name, x_label=logger.step_names[group_name])

        for idx, extra_plot in enumerate(self.extra_plots):
            ax = axes[(len(log_groups) + idx) // self.max_cols, (len(log_groups) + idx) % self.max_cols]
            extra_plot(ax, logger)

        self._after_plots(fig)

    def _default_after_subplot(self, ax: plt.Axes, group_name: str, x_label: str):
        """Add title xlabel and legend to single chart
        Args:
            ax: matplotlib Axes
            group_name: name of metrics group (eg. Accuracy, Recall)
            x_label: label of x axis (eg. epoch, iteration, batch)
        """
        ax.set_title(group_name)
        ax.set_xlabel(x_label)
        ax.legend(loc='center right')

    def _default_before_plots(self, fig: plt.Figure, axes: np.ndarray, num_of_log_groups: int) -> None:
        """Set matplotlib window properties
        Args:
            fig: matplotlib Figure
            num_of_log_groups: number of log groups
        """
        clear_output(wait=True)
        figsize_x = self.max_cols * self.cell_size[0]
        figsize_y = ((num_of_log_groups + 1) // self.max_cols + 1) * self.cell_size[1]
        fig.set_size_inches(figsize_x, figsize_y)
        if num_of_log_groups < axes.size:
            for idx, ax in enumerate(axes[-1]):
                if idx >= (num_of_log_groups + len(self.extra_plots)) % self.max_cols:
                    ax.set_visible(False)

    def _default_after_plots(self, fig: plt.Figure):
        """Set properties after charts creation
        Args:
            fig: matplotlib Figure
        """
        fig.tight_layout()
        if self.figpath is not None:
            fig.savefig(self.figpath.format(i=self.file_idx))
            self.file_idx += 1
        plt.show()

    def _draw_metric_subplot(self, ax: plt.Axes, group_logs: Dict[str, List[LogItem]], group_name: str, x_label: str):
        """
        Args:
            ax: matplotlib Axes
            group_logs: list of log items per each group name
            group_name: group name
            x_label: label for the x axis
        """
        # There used to be skip first part, but I skip it first
        if self.max_epoch is not None:
            ax.set_xlim(0, self.max_epoch)

        for name, logs in group_logs.items():
            if len(logs) > 0:
                xs = [log.step for log in logs]
                ys = [log.value for log in logs]
                ax.plot(xs, ys, label=name)

        self._after_subplot(ax, group_name, x_label)

    def _not_inline_warning(self):
        backend = matplotlib.get_backend()
        if "backend_inline" not in backend:
            warnings.warn(
                "livelossplot requires inline plots.\nYour current backend is: {}"
                "\nRun in a Jupyter environment and execute '%matplotlib inline'.".format(backend)
            )



import re
from collections import OrderedDict, defaultdict
from typing import NamedTuple, Dict, List, Pattern, Tuple, Optional, Union

# Value of metrics - for value later, we want to support numpy arrays etc
LogItem = NamedTuple('LogItem', [('step', int), ('value', float)])
COMMON_NAME_SHORTCUTS = {
    'acc': 'Accuracy',
    'nll': 'Log Loss (cost function)',
    'mse': 'Mean Squared Error',
    'loss': 'Loss'
}


class MainLogger:
    """
    Main logger - the aim of this class is to store every log from training
    Log is a float value with corresponding training engine step
    """
    def __init__(
        self,
        groups: Optional[Dict[str, List[str]]] = None,
        metric_to_name: Optional[Dict[str, str]] = None,
        current_step: int = -1,
        auto_generate_groups_if_not_available: bool = True,
        auto_generate_metric_to_name: bool = True,
        group_patterns: List[Tuple[Pattern, str]] = [
            (r'^(?!val(_|-))(.*)', 'training'),
            (r'^(val(_|-))(.*)', 'validation'),
        ],
        step_names: Union[str, Dict[str, str]] = 'epoch'
    ):
        """
        Args:
            groups: dictionary with grouped metrics for example the group 'accuracy' can contains different stages
                for example 'validation_accuracy', 'training_accuracy' etc.
            metric_to_name: transformation of metric name which can be used to display name
                we can have short name in the code (acc), but full name on charts (Accuracy)
            current_step: current step of the train loop
            auto_generate_groups_if_not_available: flag, that enable auto-creation of metric groups
                base on group patterns
            auto_generate_metric_to_name: flag, that enable auto-creation of metric long names
                base on common shortcuts
            group_patterns: you can put there regular expressions to match a few metric names with group
                and replace its name using second value
            step_names: dictionary with a name of x axis for each metrics group or one name for all metrics
        """
        self.log_history = {}
        self.groups = groups if groups is not None else {}
        self.metric_to_name = metric_to_name if metric_to_name else {}
        self.current_step = current_step
        self.auto_generate_groups = all((not groups, auto_generate_groups_if_not_available))
        self.auto_generate_metric_to_name = auto_generate_metric_to_name
        self.group_patterns = tuple((re.compile(pattern), replace_with) for pattern, replace_with in group_patterns)
        if isinstance(step_names, str):
            self.step_names = defaultdict(lambda: step_names)
        else:
            self.step_names = defaultdict(lambda: 'epoch', step_names)

    def update(self, logs: Dict[str, float], current_step: Optional[int] = None) -> None:
        """
        Args:
            logs: dictionary with metric names and values
            current_step: current step of the training loop

        Notes:
            Loop step can be controlled outside or inside main logger with autoincrement of self.current_step
        """
        if current_step is None:
            self.current_step += 1
            current_step = self.current_step
        else:
            self.current_step = current_step
        for k, v in logs.items():
            if k not in self.log_history:
                self._add_new_metric(k)
            self.log_history[k].append(LogItem(step=current_step, value=v))

    def _add_new_metric(self, metric_name: str):
        """Add empty list for a new metric and extend metric name transformations
        Args:
            metric_name: name of metric that will be added to log_history as empty list
        """
        self.log_history[metric_name] = []
        if not self.metric_to_name.get(metric_name):
            self._auto_generate_metrics_to_name(metric_name)

    def _auto_generate_metrics_to_name(self, metric_name: str):
        """The function generate transforms for metric names base on patterns
        Args:
            metric_name: name of new appended metric

        Example:
            It can create transformation from val_acc to Validation Accuracy
        """
        suffix = self._find_suffix_with_group_patterns(metric_name)
        if suffix is None and suffix != metric_name:
            return
        similar_metric_names = [m for m in self.log_history.keys() if m.endswith(suffix)]
        if len(similar_metric_names) == 1:
            return
        for name in similar_metric_names:
            new_name = name
            for pattern_to_replace, replace_with in self.group_patterns:
                new_name = re.sub(pattern_to_replace, replace_with, new_name)
            if suffix in COMMON_NAME_SHORTCUTS.keys():
                new_name = new_name.replace(suffix, COMMON_NAME_SHORTCUTS[suffix])
            self.metric_to_name[name] = new_name

    def grouped_log_history(self,
                            raw_names: bool = False,
                            raw_group_names: bool = False) -> Dict[str, Dict[str, List[LogItem]]]:
        """
        Args:
            raw_names: flag, return raw names instead of transformed by metric to name (as in update() input dictionary)
            raw_group_names: flag, return group names without transforming them with COMMON_NAME_SHORTCUTS

        Returns:
            logs grouped by metric groups - groups are passed in the class constructor

        Notes:
            method use group patterns instead of groups if they are available
        """
        if self.auto_generate_groups:
            self.groups = self._auto_generate_groups()
        ret = {}
        sorted_groups = OrderedDict(sorted(self.groups.items(), key=lambda t: t[0]))
        for group_name, names in sorted_groups.items():
            group_name = group_name if raw_group_names else COMMON_NAME_SHORTCUTS.get(group_name, group_name)
            ret[group_name] = {
                name if raw_names else self.metric_to_name.get(name, name): self.log_history[name]
                for name in names
            }
        return ret

    def _auto_generate_groups(self) -> Dict[str, List[str]]:
        """
        Returns:
            groups generated with group patterns

        Notes:
            Auto create groups base on val_ prefix - this step is skipped if groups are set
                or if group patterns are available
        """
        groups = {}
        for key in self.log_history.keys():
            abs_key = self._find_suffix_with_group_patterns(key)
            if not groups.get(abs_key):
                groups[abs_key] = []
            groups[abs_key].append(key)
        return groups

    def _find_suffix_with_group_patterns(self, metric_name: str) -> str:
        suffix = metric_name
        for pattern, _ in self.group_patterns:
            match = re.match(pattern, metric_name)
            if match:
                suffix = match.groups()[-1]
        return suffix

    def reset(self) -> None:
        """Method clears logs, groups and reset step counter"""
        self.log_history = {}
        self.groups = {}
        self.current_step = -1

    @property
    def groups(self) -> Dict[str, List[str]]:
        """groups getter"""
        return self._groups

    @groups.setter
    def groups(self, value: Dict[str, List[str]]) -> None:
        """groups setter - groups should be dictionary"""
        if value is None:
            self._groups = {}
        self._groups = value

    @property
    def log_history(self) -> Dict[str, List[LogItem]]:
        """logs getter"""
        return self._log_history

    @log_history.setter
    def log_history(self, value: Dict[str, List[LogItem]]) -> None:
        """logs setter - logs can not be overwritten - you can only reset it to empty state"""
        if len(value) > 0:
            raise RuntimeError('Cannot overwrite log history with non empty dictionary')
        self._log_history = value



from typing import Dict

from livelossplot.plot_losses import PlotLosses


class _PlotLossesCallback:
    """Base keras callback class for keras and tensorflow.keras"""
    def __init__(self, **kwargs):
        """
        Args:
            **kwargs: keyword arguments that will be passed to PlotLosses constructor
        """
        self.liveplot = PlotLosses(**kwargs)

    def on_epoch_end(self, epoch: int, logs: Dict[str, float]):
        """Send metrics to livelossplot
        Args:
            epoch: epoch number
            logs: metrics with values
        """
        self.liveplot.update(logs.copy(), epoch)
        self.liveplot.send()
